SystemOrganization addCategory: #AGV!
SystemOrganization addCategory: #'AGV-SIM'!

Object subclass: #AGV
	instanceVariableNames: 'x v ambit task target name1'
	classVariableNames: 'A AGVs Area Flane Frep Radius Simulation Vmax Vmin'
	poolDictionaries: ''
	category: 'AGV'!

!AGV commentStamp: '<historical>' prior: 0!
An AGV moves according to forces from its environment.

Instance Variables 
	ambit:			<Coll of Points>	all objects that repel me
	v:				<Point>			m/s
	x:				<Point>			m 
	task:			<Task | nil>	
	target:			<Point | nil>		Cache for target coordinate
	name1			<String>
							
Forces act between the AGV and 
i) boundary, other AGVs and other lanes - repulsive 1/r^2
ii) loadLane / unloadLane - attractive constant before visit and 
						  - repulsive after visit = i) 
Friction surrogate:
see self checkTarget and self maxV

Class variables
	AGVs 		<Dictionary of: name1 and AGV>	
	Flane 		<Number>		attractive constant force to target lane
	Frep 		<Number>		repulsive 1/r^2 force to agv & other lanes
	Vmax		<Number>		>0
	Vmin		<Number>		>0
	A			<Number>		1 > A >0
	Radius 		<Number>		of the environment
	Area 		interface object that ca be asked
				- for the ambit (environment)
				- for lane coordinates
				- to update own coordinates
				- to add or remove an AGV
	Simulation  interface object that receives
				- task reports         and sends
				- next
				- new tasks
				
Todo: =========================================
- Tests 
!

!AGV class methodsFor: 'interface-geo' stamp: 'hk 7/2/2014 20:33'!
AGVsFromArea
	Area ifNil: [ ^self ]. 
	(Area allObjectNamesAndPositionsWith: 'agv*') keysAndValuesDo: 
		[ :k :v | AGVs add: (Association key: k value: (self named: k at: v)) ]! !

!AGV class methodsFor: 'interface-simulation' stamp: 'hk 7/4/2014 16:21'!
allAGVs
	"answerds a Dictionary of the AGV names with [location, status] as value"

	^AGVs collect: [ :v | Array with: v location  with: v free  ]! !

!AGV class methodsFor: 'class initialization' stamp: 'hk 7/7/2014 17:55'!
initialize 
	AGVs := Dictionary new.
	Flane := 6 .
	Frep := 100 .
	Vmax := 5 .
	Vmin := 0.5 .
	A := 0.8 .
	Radius := 40 .! !

!AGV class methodsFor: 'instance creation' stamp: 'hk 7/4/2014 16:24'!
named: aName at: aPoint 
	^(self new) 
		location: aPoint;
		name1: aName;
		yourself! !

!AGV class methodsFor: 'interface-simulation' stamp: 'hk 7/3/2014 12:38'!
next
	AGVs do: [ :e |  e next ]! !

!AGV class methodsFor: 'class initialization' stamp: 'hk 7/2/2014 15:53'!
setArea: anArea 
	Area := anArea! !

!AGV class methodsFor: 'class initialization' stamp: 'hk 7/2/2014 15:54'!
setSimulation: aSimulation 
	Simulation := aSimulation! !

!AGV class methodsFor: 'interface-simulation' stamp: 'hk 7/2/2014 16:03'!
task: id for: agvName from: loadName to: unloadName
	(AGVs at: agvName) newTask: id from: loadName to: unloadName! !

!AGV methodsFor: 'simulation' stamp: 'hk 7/4/2014 13:56'!
checkTarget
	"loadLane / unloadLane will be reached when v=dx < vmin.
	adjust x, load or unload, switch task"
	"PRE Task isNotNil"

	(target-x) r > Vmin ifTrue: [ ^self ].			"not yet there"
	x := target.
	self updateX .
	v := 0@0. 
	task next.									"do the target action"
	self loaded ifTrue: 
		[ self flushTarget; updateLoaded ].		"unload target"
	self finished ifTrue: 
		[self reportTask; updateEmpty. 
		target := nil. task := nil ].
	! !

!AGV methodsFor: 'testing' stamp: 'hk 6/30/2014 11:24'!
empty
	^ task empty! !

!AGV methodsFor: 'testing' stamp: 'hk 7/2/2014 12:13'!
finished
	^ task finished! !

!AGV methodsFor: 'simulation' stamp: 'hk 7/4/2014 14:09'!
flushAmbit
	ambit  := Area objectLocationsWithin: Radius of:  x without:  (Array with: self targetLane with: name1). 
	ambit addAll: ( Area boundaryPointsWithin: Radius of:  x).
	 ! !

!AGV methodsFor: 'simulation' stamp: 'hk 7/2/2014 10:58'!
flushTarget
	"to be called after task progress"
	
	target := Area locationOfObjectNamed: self targetLane 
! !

!AGV methodsFor: 'simulation' stamp: 'hk 7/7/2014 14:42'!
forceFrom: aPoint
	"the repulsive force is directed from aPoint to x"
	"after load or unload the agv may be at the target,
	hence we avoid to divide by Zero"
	
	| d |
	d := x squaredDistanceTo:  aPoint.
	^d = 0 
		ifTrue: [ ^Frep asFloat * (x - aPoint) ]
		ifFalse: [ ^Frep asFloat * (x - aPoint) /  d ]
	 ! !

!AGV methodsFor: 'simulation' stamp: 'hk 7/4/2014 16:26'!
forceTo: aPoint
	"the attractive force is directed to aPoint"
	
	^Flane asFloat * (aPoint -x) / (aPoint dist: x)! !

!AGV methodsFor: 'testing' stamp: 'hk 7/2/2014 12:09'!
free
	^task = nil! !

!AGV methodsFor: 'initialize-release' stamp: 'hk 7/2/2014 14:29'!
initialize 	
	v := 0@0.
	x := 0@0.
	ambit := #()! !

!AGV methodsFor: 'testing' stamp: 'hk 6/30/2014 11:24'!
loaded
	^ task loaded! !

!AGV methodsFor: 'accessing' stamp: 'hk 7/2/2014 16:10'!
location
	^x! !

!AGV methodsFor: 'accessing' stamp: 'hk 7/4/2014 16:24'!
location: aPoint 
	x := aPoint! !

!AGV methodsFor: 'simulation' stamp: 'hk 7/7/2014 15:05'!
maxV
	"the scalar v does neither exceed vmax nor a * vd, 
	the velocity needed to reach the lane in the next step; 0.5 < a < 1"

	target ifNil: [ ^Vmax ] .
	^Vmax min: A * (target-x) r ! !

!AGV methodsFor: 'accessing' stamp: 'hk 7/2/2014 13:51'!
name1: aName 
	name1 := aName! !

!AGV methodsFor: 'simulation' stamp: 'hk 7/2/2014 16:01'!
newTask: aNumber from: a to: b
	"assign a new task"
	
	task := Task id: aNumber from: a to: b.
	self flushTarget! !

!AGV methodsFor: 'simulation' stamp: 'hk 7/7/2014 15:00'!
next
	 
	| xess |
	v := v *0.9.								"running out slowly"
	v := v + self sumOfForces.				"f = dv; m=dt =1"
	xess := v r /self maxV .
	xess > 1 ifTrue: [v := v / xess].			"not > maxV & slow to target"
	x := (x + v) asNonFractionalPoint.		" move "
	self updateX.
	task ifNotNil: [self checkTarget]		" target reached? "
	! !

!AGV methodsFor: 'testing' stamp: 'hk 7/2/2014 12:12'!
occupied
	^self free not! !

!AGV methodsFor: 'interface-simulation' stamp: 'hk 7/2/2014 16:06'!
reportTask
	Simulation ifNil: [ ^self ].
	Simulation finishedTaskNo: task id! !

!AGV methodsFor: 'simulation' stamp: 'hk 7/4/2014 16:34'!
sumOfForces
	| f |
	f := 0@0.
	self flushAmbit.
	ambit do: [ :p | f := f + (self forceFrom: p) ].
	target ifNotNil: [ f := f + (self forceTo: target) ].
	^f asNonFractionalPoint! !

!AGV methodsFor: 'accessing' stamp: 'hk 7/2/2014 12:03'!
targetLane
	self free	ifTrue: [ ^'none' ].
	self empty 	ifTrue: [ ^task loadLane].
	self loaded ifTrue: [ ^task unloadLane  ]! !

!AGV methodsFor: 'interface-geo' stamp: 'hk 7/7/2014 15:06'!
updateEmpty
	Area ifNotNil: [ Area change: name1 at: #color: to: Color yellow]! !

!AGV methodsFor: 'interface-geo' stamp: 'hk 7/2/2014 13:41'!
updateLoaded
	Area ifNotNil: [ Area change: name1 at: #color: to: Color red]! !

!AGV methodsFor: 'interface-geo' stamp: 'hk 7/2/2014 13:28'!
updateX
	Area ifNotNil: [ Area  moveObject: name1 To: x ]! !

Object subclass: #AgvConfigurator
	instanceVariableNames: 'dispatcher visualisation'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'AGV-SIM'!

!AgvConfigurator commentStamp: '<historical>' prior: 0!
An AgvConfigurator is the component that configures the agv (simulation) system: init and start-up.

Instance Variables
	dispatcher:		<AgvDispatcher>
	visualisation:	<GeoDSLDrawing>
	agv:			global class, no slot needed

Todo: =========================================
- Tests 
- UI mit start/stop  und Auftrgen (Tasks)!

!AgvConfigurator methodsFor: 'initialization' stamp: 'hk 7/4/2014 16:49'!
drawing 
	^ (GeoDSLModel new createFrom: self simpleTerminalAGVArea) drawing 		
			! !

!AgvConfigurator methodsFor: 'initialization' stamp: 'hk 7/7/2014 13:01'!
initialize 
	"AgvConfigurator new inspect"
	
	AGV initialize.		
	visualisation := self drawing.
	dispatcher := AgvDispatcher new.
	AGV setSimulation: dispatcher.
	AGV setArea: visualisation.
	AGV AGVsFromArea.
	dispatcher getAgvs .! !

!AgvConfigurator methodsFor: 'initialization' stamp: 'hk 7/3/2014 12:25'!
release 	
	AGV setSimulation: nil.
	AGV setArea: nil.
	dispatcher release
! !

!AgvConfigurator methodsFor: 'geoSpec' stamp: 'hk 7/7/2014 16:25'!
simpleTerminalAGVArea
	^ '      			
	displayIn: myGraphics size: 1000@500    	"defines the initial zoom"
	"AGVs"
		(ellipse:  agv01 size: 20@20 at: 100@150 label: agv1)
		(ellipse:  agv02 size: 20@20 at: 140@150 label: agv2)
		(ellipse:  agv03 size: 20@20 at: 160@150 label: agv3)
		(ellipse:  agv04 size: 20@20 at: 180@150 label: agv4)
		(ellipse:  agv05 size: 20@20 at: 200@150 label: agv5)
		(ellipse:  agv06 size: 20@20 at: 250@150 label: agv6)
		(ellipse:  agv07 size: 20@20 at: 280@150 label: agv7)
		(ellipse:  agv08 size: 20@20 at: 300@150 label: agv8)
		(ellipse:  agv09 size: 20@20 at: 320@150 label: agv9)
	"Lagerblock-Spuren"
		(rectangle:  laneLB1S1 size: 60@20 at: 360@30  bearing: 90   label: spur1 color: lightBlue)
		(rectangle:  laneLB1S2 size: 60@20 at: 390@30  bearing: 90   label: spur2 color: lightBlue)
		(rectangle:  laneLB1S3 size: 60@20 at: 420@30  bearing: 90   label: spur3 color: lightBlue)
		(rectangle:  laneLB1S4 size: 60@20 at: 450@30  bearing: 90   label: spur4 color: lightBlue)
		(rectangle:  laneLB2S1 size: 60@20 at: 760@30  bearing: 90   label: spur1 color: lightBlue)
		(rectangle:  laneLB2S2 size: 60@20 at: 790@30  bearing: 90   label: spur2 color: lightBlue)
		(rectangle:  laneLB2S3 size: 60@20 at: 820@30  bearing: 90   label: spur3 color: lightBlue)
		(rectangle:  laneLB2S4 size: 60@20 at: 850@30  bearing: 90   label: spur4 color: lightBlue)
	"CB-Spuren"	
		(rectangle:  laneCB1S1 size: 60@20 at: 360@240  label: spur1 color: lightGreen)
		(rectangle:  laneCB1S2 size: 60@20 at: 360@270  label: spur2 color: lightGreen)
		(rectangle:  laneCB2S1 size: 60@20 at: 460@240  label: spur1 color: lightGreen)
		(rectangle:  laneCB2S2 size: 60@20 at: 460@270  label: spur2 color: lightGreen)
		(rectangle:  laneCB3S1 size: 60@20 at: 760@240  label: spur1 color: lightGreen)
		(rectangle:  laneCB3S2 size: 60@20 at: 760@270  label: spur2 color: lightGreen)
		(rectangle:  laneCB4S1 size: 60@20 at: 860@240  label: spur1 color: lightGreen)
		(rectangle:  laneCB4S2 size: 60@20 at: 860@270  label: spur2 color: lightGreen)
	"Boundary"	
		(polygon: boundary [ 0@0  30@0  60@0  90@0  120@0  150@0  180@0  210@0  240@0  270@0  300@0 330@0  360@0  390@0  420@0  450@0  480@0  510@0  540@0  570@0  600@0  630@0  660@0  690@0  720@0  750@0  780@0  810@0  840@0  870@0  900@0  930@0  960@0  990@0 990@30  990@60  990@90  990@120  990@150  990@180  990@210  990@240  990@270   990@300  960@300  930@300  900@300  870@300  840@300  810@300  780@300  750@300  720@300  690@300  660@300  630@300  600@300  570@300  540@300  510@300  480@300  450@300  420@300  390@300  360@300  330@300  300@300  270@300  240@300  210@300  180@300  150@300  120@300  90@300  60@300  30@300  0@300 0@270   0@240   0@210   0@180   0@150   0@120   0@90   0@60   0@30   ] )
'! !

!AgvConfigurator methodsFor: 'example' stamp: 'hk 7/7/2014 13:09'!
task100
	AGV task: 100 for: 'agv06' from: 'laneCB2S2' to: 'laneLB2S3'
	! !

!AgvConfigurator methodsFor: 'example' stamp: 'hk 7/7/2014 13:10'!
task110
	AGV task: 110 for: 'agv04' from: 'laneCB1S1' to: 'laneLB1S1'
	! !

!AgvConfigurator methodsFor: 'example' stamp: 'hk 7/7/2014 14:17'!
task120
	AGV task: 120 for: 'agv05' from: 'laneCB2S2' to: 'laneLB2S3'
	! !

!AgvConfigurator methodsFor: 'example' stamp: 'hk 7/7/2014 14:17'!
task130
	AGV task: 130 for: 'agv03' from: 'laneCB1S1' to: 'laneLB1S1'
	! !

Object subclass: #AgvDispatcher
	instanceVariableNames: 'freeAgvs occAgvs freeTasks simulator running'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'AGV-SIM'!

!AgvDispatcher commentStamp: '<historical>' prior: 0!
An AgvDispatcher dispatches transport tasks to agvs.
It allows to:  
- issue a new transport task
- start or stop the simulation
the AGV component ist (as a class) globally known.
the location of the agvs will be used by sophisticated optimizers.

Instance Variables
	freeAgvs:		<Dict of: Agv-Name and location>		free
	occAgvs:		<Dict of: Agv-Name and location>		occupied
	freeTasks:		<Dict of: Task>
	simulator:		<Block>
	running: 		<Boolean>				tested by the simulator block
							

Todo: =========================================
- Tests !

!AgvDispatcher methodsFor: 'interface-agv' stamp: 'hk 7/7/2014 13:21'!
finishedTaskNo: taskID
	GrowlMorph 
		openWithLabel: 'task ', taskID printString,  ' finished' translated contents: ''! !

!AgvDispatcher methodsFor: 'interface-agv' stamp: 'hk 7/2/2014 20:27'!
getAgvs
	| agvs |
	agvs := AGV allAGVs. 
	freeAgvs := (agvs select: [ :v | v at: 2 ]) collect: [ :v | v at: 1 ].
	occAgvs := (agvs select: [ :v | (v at: 2) not ]) collect: [ :v | v at: 1 ]
	! !

!AgvDispatcher methodsFor: 'initialize-release' stamp: 'hk 7/4/2014 14:11'!
initialize 
	self stop.
	simulator := [ [true] whileTrue: [self step]] forkAt: Processor activePriority+1 .
	freeTasks := Dictionary new.
	occAgvs := Dictionary new.
	freeAgvs := Dictionary new.
	! !

!AgvDispatcher methodsFor: 'initialize-release' stamp: 'hk 7/2/2014 20:14'!
release 
	simulator terminate. 
	simulator := nil! !

!AgvDispatcher methodsFor: 'simulation' stamp: 'hk 7/2/2014 20:36'!
start 
	running := true ! !

!AgvDispatcher methodsFor: 'simulation' stamp: 'hk 7/3/2014 12:15'!
step 
	running ifTrue: 	[AGV next].
	(Delay forMilliseconds: 500) wait ! !

!AgvDispatcher methodsFor: 'simulation' stamp: 'hk 7/2/2014 20:36'!
stop 
	running := false ! !

Object subclass: #Task
	instanceVariableNames: 'loadLane unloadLane stage id'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'AGV'!

!Task commentStamp: '<historical>' prior: 0!
A Task tells an AGV where to pick up and where to drop.
initialized with #empty

Instance Variables
	loadLane:		<Symbol>				pick up
	stage:			<#empty, #loaded, #finished>
	unloadLane:	<Symbol>				drop
	id: 				<number| String>		ref

Todo: =========================================
- Tests !

!Task class methodsFor: 'instance creation' stamp: 'hk 7/2/2014 16:00'!
id: aNumber from: laneL to: laneU

	| t |
	t := self new.
	t id: aNumber.
	t loadLane: laneL.
	t unloadLane: laneU.
	^t
	! !

!Task methodsFor: 'testing' stamp: 'hk 7/2/2014 11:57'!
empty
	^stage = #empty! !

!Task methodsFor: 'testing' stamp: 'hk 7/2/2014 12:11'!
finished
	^ stage = #finished! !

!Task methodsFor: 'accessing' stamp: 'hk 7/2/2014 15:59'!
id 
	^id! !

!Task methodsFor: 'accessing' stamp: 'hk 7/2/2014 16:00'!
id: aNumber
	id := aNumber
	! !

!Task methodsFor: 'testing' stamp: 'hk 7/2/2014 12:55'!
initialize
	stage := #empty! !

!Task methodsFor: 'accessing' stamp: 'hk 6/30/2014 11:20'!
loadLane 
	^loadLane! !

!Task methodsFor: 'accessing' stamp: 'hk 7/2/2014 12:55'!
loadLane: aString 
	loadLane := aString! !

!Task methodsFor: 'testing' stamp: 'hk 6/30/2014 11:23'!
loaded
	^ stage = #loaded! !

!Task methodsFor: 'simulation' stamp: 'hk 7/7/2014 14:48'!
next    
	self empty ifTrue: [^stage := #loaded ].
	self loaded ifTrue: [^stage := #finished]! !

!Task methodsFor: 'accessing' stamp: 'hk 6/30/2014 11:20'!
unloadLane 
	^unloadLane! !

!Task methodsFor: 'accessing' stamp: 'hk 7/2/2014 12:56'!
unloadLane: aString 
	unloadLane := aString! !
AGV initialize!
